/*
 *  gunzets.c, driver for USB touchscreens made by Gunze (chipset AHL61)
 *
 *  Copyright (c) 2000 Alessandro Rubini <rubini@linux.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/init.h>
#include <linux/ctype.h>

/* usb stuff */
#include <linux/input.h>
#include <linux/usb.h>

/* miscdevice stuff */
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <asm/uaccess.h>

#define USB_VENDOR_ID_GUNZE   0x0637
#define USB_PRODUCT_ID_AHL61  0x0001

int use_input;
MODULE_PARM(use_input, "i");

#define GUNZETS_PACKET_SIZE       3 /* PS2 packet size */
#define GUNZETS_MAXPACKET_SIZE   16
#define GUNZETS_BUFFER_SIZE     480 /* a multiple of both */

#define REALPOS_MAX 16383

struct gunzets_buf {
    unsigned char *buffer;
    int head;
    int tail;
    wait_queue_head_t queue;
};


/* our device -- currently at most one is allocated */
struct gunzets_device {
    int x, y;
    unsigned char data[8]; /* aligned */
    unsigned int flags;
    struct timeval tv; /* last pen-up */
    struct urb urb;
    struct input_dev idev;
    struct gunzets_buf buffer;
} *gunzets_device;

#define GUNZETS_FLAG_PEN        0x0001 /* whether the pen/finger is down */

void gunzets_release_dev(void)
{
    struct gunzets_device *dev = gunzets_device;
    wake_up_interruptible(&dev->buffer.queue);
    if (dev && dev->buffer.buffer) kfree(dev->buffer.buffer);
    if (dev) kfree(dev);
    gunzets_device = NULL;
}   
struct gunzets_device *gunzets_alloc_dev(void)
{
    struct gunzets_device *dev;
    /* allocate and zero a new data structure for the new device */
    if (gunzets_device) return NULL; /* only one device */
    dev = kmalloc(sizeof(struct gunzets_device), GFP_KERNEL);
    if (!dev) goto out;
    memset(dev, 0, sizeof(*dev));
    dev->buffer.buffer = kmalloc(GUNZETS_BUFFER_SIZE, GFP_KERNEL);
    if (!dev->buffer.buffer)
	goto out;
    init_waitqueue_head(&dev->buffer.queue);
    return gunzets_device = dev;
 out:
    gunzets_release_dev();
    printk(KERN_ERR "gunzets: out of memory\n");
    return NULL;
}

/* fill the buffer with new data */
void gunzets_write_buffer(struct gunzets_buf *buf, char *data, int cnt)
{
    /* buffer size is a multiple of the number of bytes. Write at head */
    int nexthead = (buf->head + cnt) % GUNZETS_BUFFER_SIZE;
    if ( nexthead == buf->tail)
	return; /* buffer full */
    memcpy(buf->buffer +buf->head, data, cnt);
    buf->head = nexthead;
    wake_up_interruptible(&buf->queue);
}

/* Function in charge of receiving data */
void gunzets_irq(struct urb *urb)
{
    struct gunzets_device *dev = urb->context;
    signed char *data = dev->data;
    unsigned int flags = dev->flags, x, y;
    unsigned char out[GUNZETS_MAXPACKET_SIZE+4];

    if (urb->status != USB_ST_NOERROR) return;

    if (!(data[0]&0x80) || ((data[1] | data[2] | data[3]) &0x80)) {
	printk(KERN_WARNING "gunzets: invalid data packet\n");
	return;
    }

    if (data[0] & 0x20) dev->flags |=  GUNZETS_FLAG_PEN;
    else                dev->flags &= ~GUNZETS_FLAG_PEN;

    x = ((data[0] & 0x1f) << 7) | data[2];
    y = ((data[1] & 0x1f) << 7) | data[3];

    /* disabled messages */
    if (0 && dev->x + dev->y != 8190) {/* don't print if disconnected */
	printk(KERN_DEBUG "gunzets: data 0x%02x 0x%02x 0x%02x 0x%02x\n",
	       (int)data[0] & 0xff, data[1], data[2], data[3]);
	printk(KERN_DEBUG "gunzets: %04x %4i %04i\n",
	       flags, dev->x, dev->y);
    }

    /* write to device buffer in PS2 format */
    x = x/4; y = y/4;
#if 0
    printk("%4i,%4i (%i)\n", x, y, ((dev->flags & GUNZETS_FLAG_PEN) ? 1 : 0));
#endif
    out[0] = 0x80 | ((x >> 3) & 0x7f);   /* X9-3 */
    out[1] = 0x80 | ((x & 0x07) << 4) | ((y >> 6) & 0x0f);   /* X2-0, Y9-6 */
    out[2] = 0x00 | (y & 0x3f) | ((dev->flags & GUNZETS_FLAG_PEN) ? 0x40 : 0);   /* Y5-0, T/~R */
    gunzets_write_buffer(&dev->buffer, out, GUNZETS_PACKET_SIZE);
    if (!(dev->flags & GUNZETS_FLAG_PEN)) dev->x = dev->y = -1; /* invalid */
}


/*
 * These two callbacks are invoked when an USB device is detached or attached
 * to the bus
 */

void gunzets_disconnect(struct usb_device *udev, void *clientdata)
{
    /* the clientdata is the gunzets_device we passed originally */
    struct gunzets_device *dev = clientdata;

    /* remove the URB, remove the input device, free memory */
    usb_unlink_urb(&dev->urb);
    if (use_input) {
	input_unregister_device(&dev->idev);
    }
    gunzets_release_dev();
    printk(KERN_INFO "gunzets: touchscreen disconnected\n");
    MOD_DEC_USE_COUNT;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
/* version 2.2.18 has USB support but the probe prototype is different */
    void *gunzets_probe(struct usb_device *udev, unsigned int ifnum)
#else
    void *gunzets_probe(struct usb_device *udev, unsigned int ifnum,
			const struct usb_device_id *id)
#endif
{
    struct usb_interface_descriptor *interface;
    struct usb_endpoint_descriptor *endpoint;
    struct gunzets_device *dev;

    if (udev->descriptor.idVendor  != USB_VENDOR_ID_GUNZE)  return NULL;
    if (udev->descriptor.idProduct != USB_PRODUCT_ID_AHL61) return NULL;

    /* it looks like this is the right device, but be paranoid */
    if (udev->descriptor.bNumConfigurations != 1) return NULL;
    interface = udev->config[0].interface[ifnum].altsetting;

    if (interface->bNumEndpoints != 1) return NULL;
    endpoint = interface->endpoint;

    if (!(endpoint->bEndpointAddress & 0x80)) return NULL; /* not input? */
    if ((endpoint->bmAttributes & 3) != 3) return NULL; /* not interrupt tr. */

    if (!(dev = gunzets_alloc_dev())) return NULL;

    usb_set_protocol(udev, interface->bInterfaceNumber, 0);
    usb_set_idle(udev, interface->bInterfaceNumber, 0, 0);

    /* fill the URB data structure using the FILL_INT_URB macro */
    {
	int pipe = usb_rcvintpipe(udev, endpoint->bEndpointAddress);
	int maxp = usb_maxpacket(udev, pipe, usb_pipeout(pipe));
	
	FILL_INT_URB(&dev->urb, udev, pipe, dev->data,
		     maxp > 4 ? 4 : maxp, gunzets_irq, dev,
		     endpoint->bInterval);
    }

    /* register the URB within the USB subsystem */
    if (usb_submit_urb(&dev->urb)) {
	kfree(dev);
	return NULL;
    }

    if (use_input) {
	/* tell the features of this input device */
	dev->idev.evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
	set_bit(ABS_X,     dev->idev.absbit);
	set_bit(ABS_Y,     dev->idev.absbit);
	set_bit(BTN_LEFT,  dev->idev.keybit);
      
	/* and register the input device itself */
	input_register_device(&dev->idev);
    }
    /* announce yourself */
    printk(KERN_INFO "gunzets: device connected\n");

    /* and return the new structure */
    MOD_INC_USE_COUNT;
    return dev;
}


/*
 * The callbacks are registered within the USB subsystem using the
 * usb_driver data structure
 */
struct usb_driver gunzets_usb_driver = {
    name:        "gunzets",
    probe:       gunzets_probe,
    disconnect:  gunzets_disconnect,
};


/*
 * Misc device stuff 
 */

ssize_t gunzets_read_from_buffer(struct file *filp , char *buf, size_t count,
				 struct gunzets_buf *tsbuf)
{
    int i;

    while (tsbuf->head == tsbuf->tail) {
	if (filp->f_flags & O_NONBLOCK)
	    return -EAGAIN;
	interruptible_sleep_on(&tsbuf->queue);
	if (signal_pending(current)) /* a signal arrived */
	    return -ERESTARTSYS; /* tell the fs layer to handle it */
    }
    i = tsbuf->head - tsbuf->tail;
    if (i<0) i = GUNZETS_BUFFER_SIZE - tsbuf->tail;
    if (count > i) count = i;
    copy_to_user(buf, tsbuf->buffer + tsbuf->tail, count);
    tsbuf->tail = (tsbuf->tail + count) % GUNZETS_BUFFER_SIZE;
    return count;
}

ssize_t gunzets_read(struct file *filp , char *buf, size_t count,
		     loff_t *offp)
{
    struct gunzets_buf *tsbuf;

    if (!gunzets_device) return 0; /* disconnected */
    tsbuf = filp->private_data;
    return gunzets_read_from_buffer(filp, buf, count, tsbuf);
}

unsigned int gunzets_poll(struct file *filp, struct poll_table_struct *table)
{
    int ret = 0;
    struct gunzets_buf *buf;

    if (!gunzets_device) return POLLIN | POLLHUP;

    buf = filp->private_data;
    if (buf->head != buf->tail)
	ret |= POLLRDNORM | POLLIN;
    else
	poll_wait(filp, &buf->queue, table);
    return ret;
}

int gunzets_open(struct inode *inode, struct file *filp)
{
    struct gunzets_buf *buf;

    /* write is always allowed, reading only if the device is plugged */
    if (!(filp->f_mode & FMODE_READ)) {
	MOD_INC_USE_COUNT;
	return 0;
    }
    if (!gunzets_device) return -ENODEV;
    
    buf = &gunzets_device->buffer;

    /* reset the buffer just opened: forget old data */
    buf->head = buf->tail = 0;
    filp->private_data = buf;
    MOD_INC_USE_COUNT;
    return 0;
}

int gunzets_release(struct inode *inode, struct file *filp)
{
    MOD_DEC_USE_COUNT;
    return 0;
}

struct file_operations gunzets_file_operations = {
	read:     gunzets_read,
	poll:     gunzets_poll,
	open:     gunzets_open,
	release:  gunzets_release,
};


struct miscdevice gunzets_misc = {
    name:       "gunzets",
    minor:      MISC_DYNAMIC_MINOR,
    fops:       &gunzets_file_operations,
};


/*
 * Functions called at module load and unload time. Keep track of what
 * is registered, as the unload function is used to undo partial
 * inizialitazion
 */

static int init_flags;
#define DONE_USB          0x01
#define DONE_MISC         0x02

void gunzets_exit(void)
{
    if (init_flags & DONE_USB)
	usb_deregister(&gunzets_usb_driver);
    if (init_flags & DONE_MISC)
	misc_deregister(&gunzets_misc);
    init_flags = 0; /* overconvervative */
    gunzets_misc.minor = MISC_DYNAMIC_MINOR;
}


int gunzets_init(void)
{
    int error;

    error = usb_register(&gunzets_usb_driver);
    if (use_input)
	return error;

    /* not input: register the misc devices as well */
    if (!error) {
	init_flags |= DONE_USB;
	error = misc_register(&gunzets_misc);
    } else {
	gunzets_exit();
	return error;
    }
    return 0;
}


module_init(gunzets_init);
module_exit(gunzets_exit);



